# -*- coding: utf-8 -*- #
"""
Time                2023/9/4 16:23
Author:             mingfeng (SunnyQjm)
Email               mfeng@linux.alibaba.com
File                staragent.py.py
Description:
"""
import re
import hmac
import time
import aiohttp
import asyncio
from hashlib import sha1
from asyncer import syncify
from keycenter import kc_crypto_service as kcs
from keycenter import kc_key_sync as kks
from .base import BaseChannel, ChannelResult, ChannelException, ChannelCode

DEFAULT_EXEC_TIMEOUT = 600000  # 默认执行超时时间 60s

KEY_CENTER_KEY_NAME = "sysom_aone_key"
KEY_CENTER_HOST = "xxx"
KEY_CENTER_PUBLISH_NUMBER = "xxx"

# # daily
# KEY_CENTER_HOST = "https://daily-keycenter.alibaba.net/keycenter"
# KEY_CENTER_PUBLISH_NUMBER = "fb412e53d254492384b6db716dd6576e"

# # publish
# KEY_CENTER_HOST = "http://keycenter-service.alibaba-inc.com/keycenter"
# KEY_CENTER_PUBLISH_NUMBER = "c546dac321c744cdadaa9bc576001987"


STARAGENT_ENCRYPTED_SECRET = "xxx"
STARAGENT_ENCRYPTED_KEY = "xxx"
STARAGENT_BASE_URL = "xxx"

# # test
# STARAGENT_ENCRYPTED_SECRET = "i64Zkd/GzipQV+a2NYBH3n5dZgmvdNxffvgeV4Uh79eewv5Io9P23vwPTjUh7Uob"
# STARAGENT_ENCRYPTED_KEY = "xa4+zDTtLK63VFGLJPc1CSgP5IdX9ENU5iqSh5AJL/Cjoy9tQM4tlXVnJOtfAK5I"
# STARAGENT_BASE_URL = "http://inc.agent.alibaba-inc.com"

# # publish
# STARAGENT_ENCRYPTED_SECRET = "58g9cm7vfd/3cWIYZ+Izpsvhc+vuEvurPXdl0zP5X0JoCOMDbv56Yrk17IXYGj9C"
# STARAGENT_ENCRYPTED_KEY = "wcNPEE3DBKXQo6AkJJCUkoF1Iwz0hwVnUJgoVv/+FUwUkMcwOhWD9wO6EfZFX/hx"
# STARAGENT_BASE_URL = "http://staragent.alibaba-inc.com"

STARAGENT_TASK_URL = f"{STARAGENT_BASE_URL}/api/task"


class Channel(BaseChannel):
    """
    Cloud Assist channel, required sts token
    """

    def __init__(self, **kwargs) -> None:
        # Extract required params
        missing_param_list = []
        for require_param in ["instance", "command", "instance_type",
                              "service_name"]:
            if require_param not in kwargs:
                missing_param_list.append(require_param)
        if len(missing_param_list) > 0:
            raise ChannelException(
                f"StarAgent: Missing parameters: "
                f"{', '.join(missing_param_list)}",
                code=ChannelCode.REQUEST_PARAM_ERROR.value
            )

        # 目标实例，可以通过 sn/ip/hostname 区分
        self._instance_type = kwargs.get("instance_type", "")
        if self._instance_type not in ["servicetag", "ip", "hostname"]:
            raise ChannelException(
                f'StarAgent: instance_type must be one of '
                f'["servicetag", "ip", "hostname"]'
            )
        self._instance = kwargs["instance"]
        self._command = kwargs.get("command", None)
        self._service_name = kwargs.get("service_name", "")

        # Initial kc, for decrypt
        kks.initialize(KEY_CENTER_HOST, KEY_CENTER_PUBLISH_NUMBER)

        self.staragent_secret = kcs.decrypt(KEY_CENTER_KEY_NAME,
                                            STARAGENT_ENCRYPTED_SECRET)
        self.staragent_key = kcs.decrypt(KEY_CENTER_KEY_NAME,
                                         STARAGENT_ENCRYPTED_KEY)

    @staticmethod
    def initial(**kwargs) -> ChannelResult:
        return ChannelResult()

    @staticmethod
    async def initial_async(**kwargs) -> ChannelResult:
        return ChannelResult()

    def run_command(self, **kwargs) -> ChannelResult:
        """Execute command on remote machine

        Args:
            **kwargs:

        Returns:

        """
        return syncify(self.run_command_async, raise_sync_error=False)(
            **kwargs
        )

    async def _run_command_by_staragent(self, params: dict):
        def generate_params_sign():
            """
            params = {
                'ip': IP
                'key': KEY,
                'exeurl': "cmd://" + cmd,
                'timestamp': int(time.time())
            }

            :param params:
            :return:
            """
            sign_str = ''
            for k in sorted(params.keys()):
                sign_str += k
                sign_str += str(params[k])  # 参数排序

            return hmac.new(self.staragent_secret.encode(),
                            sign_str.encode(), sha1).digest().hex()  # sign加密

        sign = generate_params_sign()
        params["sign"] = sign
        async with aiohttp.ClientSession() as session:
            async with session.post(STARAGENT_TASK_URL,
                                    params=params) as resp:
                return await resp.json()

    async def _do_run_command(self, command: str, **kwargs) -> ChannelResult:
        """

        {
            "UID": string,            // 任务的UID
            "SUCCESS": true | false,  // 若为false，则通过ERRORTYPE、ERRORCODE以及ERRORMSG查看错误原因
            "ERRORTYPE": string,      // 错误类型,分为: SERVER_ERROR(机器问题), STARAGENT_ERROR(staragent系统问题), CALL_ERROR(用户调用问题), NET_ERROR(网络问题)
            "ERRORCODE": string       // 错误编码
            "ERRORMSG": string,      // 错误原因描述，success为true时，可以不关心ERRORMSG字段
            "ERRORURL": string,      // 错误排查链接，根据该链接可以自助解决一些错误场景
            "JOBNAME": string,       //JOBNAME
            "JOBRESULT": string       // 任务脚本执行结果，即脚本的标准输出内容 (长度不能超过一定大小，默认限制stdout 1M stderr 64k)
            "SERVICETAG | HOSTNAME | IP": string       // 根据任务执行时提供的参数，返回其一
        }
        Args:
            command:
            **kwargs:

        Returns:

        """

        def get_trim_match_obj_value(match_obj, idx: int):
            value = match_obj.group(idx)
            if value is None:
                return ""
            else:
                return value.strip()

        def convert_cmd(origin_cmd: str, service_name: str):
            """
            Convert origin diagnosis cmd to staragent SysOM plugin cmd
            Args:
                origin_cmd:

            Returns:

            """
            match_obj = re.match(
                r'(.*)sysak (-g )*(memgraph|podmem|iofsstat|iosdiag|rtrace|'
                r'loadtask|oomcheck|ossre_client|pingtrace|schedmoni)(.*?)('
                r'>>|&&|>|&)(.*)',
                origin_cmd, re.I)
            # 1 => sysak 之前的命令
            # 2 => -g or None
            # 3 => sysak subcommand (memgraph|podmem|iofsstat|iosdiag|rtrace|loadtask|oomcheck|ossre_client|pingtrace|schedmoni)
            # 4 => sysak subcommand 参数
            # 5 => >> | && | > | &
            # 6 => sysak 之后的命令
            if match_obj:
                # sysak_pre = get_trim_match_obj_value(match_obj, 1)
                sysak_g = get_trim_match_obj_value(match_obj, 2)
                sysak_sub_cmd = get_trim_match_obj_value(match_obj, 3)
                sysak_params = get_trim_match_obj_value(match_obj, 4)
                # sysak_sep = get_trim_match_obj_value(match_obj, 5)
                # sysak_post = get_trim_match_obj_value(match_obj, 6)

                sub_cmd = f"{'' if not sysak_g else ' ' + sysak_g} {sysak_sub_cmd} {sysak_params}"

                return f"plugin://sysom/sysak_cmd.sh fragment {service_name} {sub_cmd}"
            else:
                raise ChannelException(
                    message=f"StaragentChannel: Can not convert '{origin_cmd}'"
                            f" to staragent SysOM plugin cmd")

        async def do_fragment():
            params = {
                self._instance_type: self._instance,
                'key': self.staragent_key,
                'exeurl': convert_cmd(command, self._service_name),
                # 'exeurl': f"plugin://sysom/sysak_cmd.sh loadtask loadtask -s -g",
                'timestamp': int(time.time())
            }
            task_result = await self._run_command_by_staragent(params)
            if not task_result.get("SUCCESS", False):
                error_type = task_result.get("ERRORTYPE")
                error_code = task_result.get("ERRORCODE")
                error_msg = task_result.get("ERRORMSG")
                error_url = task_result.get("ERRORURL")
                err_msg = f"{error_type}({error_code})" \
                          f": {error_msg} => {error_url}"
                raise ChannelException(
                    message=err_msg,
                    code=ChannelCode.CHANNEL_EXEC_FAILED.value
                )
            else:
                job_result = task_result.get("JOBRESULT", "")
                if not job_result.startswith("/tmp/sysom"):
                    raise ChannelException(
                        message=f"Staragent: Execute command failed "
                                f"=> {job_result}"
                    )
                return job_result.strip().split("\n")

        async def fetch_chunk(chunk_path: str, finished: bool):
            params = {
                self._instance_type: self._instance,
                'key': self.staragent_key,
                'exeurl': f"plugin://sysom/sysak_cmd.sh fetch_chunk"
                          f" {chunk_path} {1 if finished else 0}",
                'timestamp': int(time.time())
            }
            task_result = await self._run_command_by_staragent(params)
            if not task_result.get("SUCCESS", False):
                error_type = task_result.get("ERRORTYPE")
                error_code = task_result.get("ERRORCODE")
                error_msg = task_result.get("ERRORMSG")
                error_url = task_result.get("ERRORURL")
                err_msg = f"{error_type}({error_code})" \
                          f": {error_msg} => {error_url}"
                raise ChannelException(
                    message=err_msg,
                    code=ChannelCode.CHANNEL_EXEC_FAILED.value
                )
            else:
                job_result = task_result.get("JOBRESULT", "")
                return job_result

        channel_result = ChannelResult(
            code=ChannelCode.SERVER_ERROR.value,
            err_msg="Staragent: Unknown error while invoke command",
            result="Staragent: Unknown error while invoke command"
        )

        try:
            # 1. Invoke diagnosis cmd, and do fragment
            chunks = await do_fragment()
            print(chunks)

            # 2. Limit Max chunk num
            if len(chunks) > 10:
                raise ChannelException(
                    f"StaragentChannel: Max limit {950 * 10 / 1024}MB")

            # 3. Fetch each chunk by parallel
            tasks = []
            for idx, chunk in enumerate(chunks):
                tasks.append(
                    fetch_chunk(chunk, idx == len(chunks) - 1)
                )
            res = await asyncio.gather(*tasks)

            channel_result.code = ChannelCode.SUCCESS.value
            channel_result.result = "".join(res)
            channel_result.err_msg = ""
        except ChannelException as ce:
            raise ce
        except Exception as exc:
            raise ChannelException(
                (f"Staragent: Connect failed (instance = {self._instance}) "
                 f"=> {str(exc)}"),
                code=ChannelCode.CHANNEL_CONNECT_FAILED.value,
                summary=(f"Staragent: Connect failed "
                         f"(instance = {self._instance})")
            ) from exc
        return channel_result

    async def run_command_async(self, **kwargs) -> ChannelResult:
        """Execute command on remote machine by async

        Args:
            **kwargs:

        Returns:

        """
        return await self._do_run_command(self._command, **kwargs)